-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local im  = ui_imgui

local C = {}

C.name = 'Vehicle Collision Tracking'
C.color = ui_flowgraph_editor.nodeColors.vehicle
C.icon = ui_flowgraph_editor.nodeIcons.vehicle

C.description = 'Tracks collision events between a main vehicle and a target vehicle, with some extra parameters.'
C.category = 'once_p_duration'
C.tmpSecondPassFlag = true
C.pinSchema = {

  { dir = 'in', type = 'number', name = 'vehId', description = 'Vehicle id.' },
  { dir = 'in', type = 'number', name = 'targetVehId', description = 'Target vehicle id.' },
  { dir = 'in', type = 'number', name = 'hitDuration', hidden = true, default = 0, description = '(Optional) Duration needed for a collision to count.' },
  { dir = 'in', type = 'number', name = 'reboundDuration', hidden = true, default = 1, description = '(Optional) Duration needed after a collision for it to count again; cooldown time' },
  { dir = 'in', type = 'bool', name = 'highPrecision', hidden = true, description = '(Optional) If true, bounding box corner detection is also used to improve detection accuracy.' },

  { dir = 'out', type = 'flow', name = 'touching', description = 'Outflow while vehicles are touching.' },
  { dir = 'out', type = 'flow', name = 'hit', impulse = true, description = 'Outflow for the instant a hit is counted.' },
  { dir = 'out', type = 'number', name = 'hitCount', description = 'Number of independent collisions.' },
  { dir = 'out', type = 'bool', name = 'touchBool', hidden = true, description = 'Is true while vehicles are touching.' }
}
C.tags = {'collision', 'collide', 'hit'}

function C:init()
  self:onNodeReset()
end

function C:_executionStopped()
  self:onNodeReset()
end

function C:onNodeReset()
  self.hitCount = 0
  self.hitTimer = 0
  self.reboundTimer = math.huge
  self.hitFlag = false
end

local bbPoints = {0, 3, 7, 4}
function C:checkBounds(vehId, targetVehId)
  local bbA = be:getObjectByID(vehId):getSpawnWorldOOBB()
  local bbB = be:getObjectByID(targetVehId):getSpawnWorldOOBB()
  local centerA = bbB:getCenter()
  local centerB = bbB:getCenter()
  local inside = true

  if math.abs(centerA.z - centerB.z) <= bbA:getHalfExtents().z + bbB:getHalfExtents().z then
    for _, a in ipairs(bbPoints) do
      inside = true
      for i, b in ipairs(bbPoints) do
        local p = i < 4 and i + 1 or 1
        local xnorm1, xnorm2 = closestLinePoints(bbA:getPoint(a):z0(), centerB:z0(), bbB:getPoint(b):z0(), bbB:getPoint(bbPoints[p]):z0())
        if xnorm1 > 0 and xnorm1 < 1 and xnorm2 > 0 and xnorm2 < 1 then
          inside = false
        end
      end
      if inside then break end
    end
  end

  return inside
end

function C:work()
  self.pinOut.touching.value = false
  self.pinOut.touchBool.value = false
  self.pinOut.hit.value = false

  local hitDuration = self.pinIn.hitDuration.value or 0
  local reboundDuration = self.pinIn.reboundDuration.value or 1

  if map.objects[self.pinIn.vehId.value] and map.objects[self.pinIn.targetVehId.value] then
    local cols = map.objects[self.pinIn.vehId.value].objectCollisions
    local inside = true

    if self.pinIn.highPrecision.value then
      inside = self:checkBounds(self.pinIn.vehId.value, self.pinIn.targetVehId.value)
    end

    if cols[self.pinIn.targetVehId.value] and inside then
      if self.hitTimer >= hitDuration then
        self.pinOut.touching.value = true
        self.pinOut.touchBool.value = true

        if not self.hitFlag and self.reboundTimer >= reboundDuration then
          self.pinOut.hit.value = true
          self.hitCount = self.hitCount + 1
          self.reboundTimer = 0
          self.hitFlag = true
        end
      end

      self.hitTimer = self.hitTimer + self.mgr.dtSim
    else
      self.hitTimer = 0
      if not self.pinIn.highPrecision.value then
        self.hitFlag = false
      end
    end

    if not inside then
      self.hitFlag = false
    end

    if self.reboundTimer < reboundDuration then
      self.reboundTimer = self.reboundTimer + self.mgr.dtSim
    end
  end

  self.pinOut.hitCount.value = self.hitCount
end

return _flowgraph_createNode(C)
